@@ -24,7 +24,7 @@
  */
 
 #include "bsdtar_platform.h"
-__FBSDID("$FreeBSD: src/usr.bin/tar/write.c,v 1.70 2008/05/26 17:10:10 kientzle Exp $");
+__FBSDID("$FreeBSD: src/usr.bin/tar/write.c,v 1.79 2008/11/27 05:49:52 kientzle Exp $");
 
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
@@ -44,7 +44,8 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/write.c,v 1.70 2008/05/26 17:10:10 kientzle
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
 #endif
-#ifdef HAVE_EXT2FS_EXT2_FS_H
+#if defined(HAVE_EXT2FS_EXT2_FS_H) && !defined(__CYGWIN__)
+/* This header exists but is broken on Cygwin. */
 #include <ext2fs/ext2_fs.h>
 #endif
 #ifdef HAVE_FCNTL_H
@@ -79,6 +80,9 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/write.c,v 1.70 2008/05/26 17:10:10 kientzle
 #include "bsdtar.h"
 #include "tree.h"
 
+/* Size of buffer for holding file data prior to writing. */
+#define FILEDATABUFLEN	65536
+
 /* Fixed size of uname/gname caches. */
 #define	name_cache_size 101
 
@@ -137,9 +141,9 @@ static void		 write_entry(struct bsdtar *, struct archive *,
 			     const struct stat *, const char *pathname,
 			     const char *accpath);
 static void		 write_entry_backend(struct bsdtar *, struct archive *,
-			     struct archive_entry *, int);
+			     struct archive_entry *);
 static int		 write_file_data(struct bsdtar *, struct archive *,
-			     int fd);
+			     struct archive_entry *, int fd);
 static void		 write_hierarchy(struct bsdtar *, struct archive *,
 			     const char *);
 
@@ -152,9 +156,6 @@ tar_mode_c(struct bsdtar *bsdtar)
 	if (*bsdtar->argv == NULL && bsdtar->names_from_file == NULL)
 		bsdtar_errc(bsdtar, 1, 0, "no files or directories specified");
 
-	/* We want to catch SIGINFO and SIGUSR1. */
-	siginfo_init(bsdtar);
-
 	a = archive_write_new();
 
 	/* Support any format that the library supports. */
@@ -216,16 +217,6 @@ tar_mode_c(struct bsdtar *bsdtar)
 		bsdtar_errc(bsdtar, 1, 0, archive_error_string(a));
 
 	write_archive(a, bsdtar);
-
-	if (bsdtar->option_totals) {
-		fprintf(stderr, "Total bytes written: " BSDTAR_FILESIZE_PRINTF "\n",
-		    (BSDTAR_FILESIZE_TYPE)archive_position_compressed(a));
-	}
-
-	archive_write_finish(a);
-
-	/* Restore old SIGINFO + SIGUSR1 handlers. */
-	siginfo_done(bsdtar);
 }
 
 /*
@@ -244,9 +235,6 @@ tar_mode_r(struct bsdtar *bsdtar)
 	/* Sanity-test some arguments and the file. */
 	test_for_append(bsdtar);
 
-	/* We want to catch SIGINFO and SIGUSR1. */
-	siginfo_init(bsdtar);
-
 	format = ARCHIVE_FORMAT_TAR_PAX_RESTRICTED;
 
 	bsdtar->fd = open(bsdtar->filename, O_RDWR | O_CREAT, 0666);
@@ -317,12 +305,6 @@ tar_mode_r(struct bsdtar *bsdtar)
 
 	write_archive(a, bsdtar); /* XXX check return val XXX */
 
-	if (bsdtar->option_totals) {
-		fprintf(stderr, "Total bytes written: " BSDTAR_FILESIZE_PRINTF "\n",
-		    (BSDTAR_FILESIZE_TYPE)archive_position_compressed(a));
-	}
-
-	archive_write_finish(a);
 	close(bsdtar->fd);
 	bsdtar->fd = -1;
 }
@@ -345,9 +327,6 @@ tar_mode_u(struct bsdtar *bsdtar)
 	/* Sanity-test some arguments and the file. */
 	test_for_append(bsdtar);
 
-	/* We want to catch SIGINFO and SIGUSR1. */
-	siginfo_init(bsdtar);
-
 	bsdtar->fd = open(bsdtar->filename, O_RDWR);
 	if (bsdtar->fd < 0)
 		bsdtar_errc(bsdtar, 1, errno,
@@ -406,12 +385,6 @@ tar_mode_u(struct bsdtar *bsdtar)
 
 	write_archive(a, bsdtar);
 
-	if (bsdtar->option_totals) {
-		fprintf(stderr, "Total bytes written: " BSDTAR_FILESIZE_PRINTF "\n",
-		    (BSDTAR_FILESIZE_TYPE)archive_position_compressed(a));
-	}
-
-	archive_write_finish(a);
 	close(bsdtar->fd);
 	bsdtar->fd = -1;
 
@@ -434,6 +407,13 @@ write_archive(struct archive *a, struct bsdtar *bsdtar)
 	const char *arg;
 	struct archive_entry *entry, *sparse_entry;
 
+	/* We want to catch SIGINFO and SIGUSR1. */
+	siginfo_init(bsdtar);
+
+	/* Allocate a buffer for file data. */
+	if ((bsdtar->buff = malloc(FILEDATABUFLEN)) == NULL)
+		bsdtar_errc(bsdtar, 1, 0, "cannot allocate memory");
+
 	if ((bsdtar->resolver = archive_entry_linkresolver_new()) == NULL)
 		bsdtar_errc(bsdtar, 1, 0, "cannot create link resolver");
 	archive_entry_linkresolver_set_strategy(bsdtar->resolver,
@@ -453,7 +433,7 @@ write_archive(struct archive *a, struct bsdtar *bsdtar)
 					bsdtar_warnc(bsdtar, 1, 0,
 					    "Missing argument for -C");
 					bsdtar->return_value = 1;
-					return;
+					goto cleanup;
 				}
 			}
 			set_chdir(bsdtar, arg);
@@ -473,8 +453,7 @@ write_archive(struct archive *a, struct bsdtar *bsdtar)
 	entry = NULL;
 	archive_entry_linkify(bsdtar->resolver, &entry, &sparse_entry);
 	while (entry != NULL) {
-		int fd = -1;
-		write_entry_backend(bsdtar, a, entry, fd);
+		write_entry_backend(bsdtar, a, entry);
 		archive_entry_free(entry);
 		entry = NULL;
 		archive_entry_linkify(bsdtar->resolver, &entry, &sparse_entry);
@@ -485,6 +464,20 @@ write_archive(struct archive *a, struct bsdtar *bsdtar)
 		bsdtar_warnc(bsdtar, 0, "%s", archive_error_string(a));
 		bsdtar->return_value = 1;
 	}
+
+cleanup:
+	/* Free file data buffer. */
+	free(bsdtar->buff);
+
+	if (bsdtar->option_totals) {
+		fprintf(stderr, "Total bytes written: " BSDTAR_FILESIZE_PRINTF "\n",
+		    (BSDTAR_FILESIZE_TYPE)archive_position_compressed(a));
+	}
+
+	archive_write_finish(a);
+
+	/* Restore old SIGINFO + SIGUSR1 handlers. */
+	siginfo_done(bsdtar);
 }
 
 /*
@@ -616,22 +609,23 @@ append_archive(struct bsdtar *bsdtar, struct archive *a, struct archive *ina)
 static int
 copy_file_data(struct bsdtar *bsdtar, struct archive *a, struct archive *ina)
 {
-	char	buff[64*1024];
 	ssize_t	bytes_read;
 	ssize_t	bytes_written;
 	off_t	progress = 0;
 
-	bytes_read = archive_read_data(ina, buff, sizeof(buff));
+	bytes_read = archive_read_data(ina, bsdtar->buff, FILEDATABUFLEN);
 	while (bytes_read > 0) {
 		siginfo_printinfo(bsdtar, progress);
 
-		bytes_written = archive_write_data(a, buff, bytes_read);
+		bytes_written = archive_write_data(a, bsdtar->buff,
+		    bytes_read);
 		if (bytes_written < bytes_read) {
 			bsdtar_warnc(bsdtar, 0, "%s", archive_error_string(a));
 			return (-1);
 		}
 		progress += bytes_written;
-		bytes_read = archive_read_data(ina, buff, sizeof(buff));
+		bytes_read = archive_read_data(ina, bsdtar->buff,
+		    FILEDATABUFLEN);
 	}
 
 	return (0);
@@ -648,10 +642,6 @@ write_hierarchy(struct bsdtar *bsdtar, struct archive *a, const char *path)
 	dev_t first_dev = 0;
 	int dev_recorded = 0;
 	int tree_ret;
-#ifdef __linux
-	int	 fd, r;
-	unsigned long fflags;
-#endif
 
 	tree = tree_open(path);
 
@@ -666,8 +656,15 @@ write_hierarchy(struct bsdtar *bsdtar, struct archive *a, const char *path)
 		const struct stat *st = NULL, *lst = NULL;
 		int descend;
 
-		if (tree_ret == TREE_ERROR_DIR)
-			bsdtar_warnc(bsdtar, errno, "%s: Couldn't visit directory", name);
+		if (tree_ret == TREE_ERROR_FATAL)
+			bsdtar_errc(bsdtar, 1, tree_errno(tree),
+			    "%s: Unable to continue traversing directory tree",
+			    name);
+		if (tree_ret == TREE_ERROR_DIR) {
+			bsdtar_warnc(bsdtar, errno,
+			    "%s: Couldn't visit directory", name);
+			bsdtar->return_value = 1;
+		}
 		if (tree_ret != TREE_REGULAR)
 			continue;
 		lst = tree_current_lstat(tree);
@@ -714,23 +711,25 @@ write_hierarchy(struct bsdtar *bsdtar, struct archive *a, const char *path)
 		 * If this file/dir is flagged "nodump" and we're
 		 * honoring such flags, skip this file/dir.
 		 */
-#ifdef HAVE_CHFLAGS
+#ifdef HAVE_STRUCT_STAT_ST_FLAGS
+		/* BSD systems store flags in struct stat */
 		if (bsdtar->option_honor_nodump &&
 		    (lst->st_flags & UF_NODUMP))
 			continue;
 #endif
 
-#ifdef __linux
-		/*
-		 * Linux has a nodump flag too but to read it
-		 * we have to open() the file/dir and do an ioctl on it...
-		 */
-		if (bsdtar->option_honor_nodump &&
-		    ((fd = open(name, O_RDONLY|O_NONBLOCK)) >= 0) &&
-		    ((r = ioctl(fd, EXT2_IOC_GETFLAGS, &fflags)),
-			close(fd), r) >= 0 &&
-		    (fflags & EXT2_NODUMP_FL))
-			continue;
+#if defined(EXT2_IOC_GETFLAGS) && defined(EXT2_NODUMP_FL)
+		/* Linux uses ioctl to read flags. */
+		if (bsdtar->option_honor_nodump) {
+			int fd = open(name, O_RDONLY | O_NONBLOCK);
+			if (fd >= 0) {
+				unsigned long fflags;
+				int r = ioctl(fd, EXT2_IOC_GETFLAGS, &fflags);
+				close(fd);
+				if (r >= 0 && (fflags & EXT2_NODUMP_FL))
+					continue;
+			}
+		}
 #endif
 
 		/*
@@ -798,11 +797,12 @@ write_hierarchy(struct bsdtar *bsdtar, struct archive *a, const char *path)
  */
 static void
 write_entry_backend(struct bsdtar *bsdtar, struct archive *a,
-    struct archive_entry *entry, int fd)
+    struct archive_entry *entry)
 {
+	int fd = -1;
 	int e;
 
-	if (fd == -1 && archive_entry_size(entry) > 0) {
+	if (archive_entry_size(entry) > 0) {
 		const char *pathname = archive_entry_sourcepath(entry);
 		fd = open(pathname, O_RDONLY);
 		if (fd == -1) {
@@ -835,10 +835,16 @@ write_entry_backend(struct bsdtar *bsdtar, struct archive *a,
 	 * that case, just skip the write.
 	 */
 	if (e >= ARCHIVE_WARN && fd >= 0 && archive_entry_size(entry) > 0) {
-		if (write_file_data(bsdtar, a, fd))
+		if (write_file_data(bsdtar, a, entry, fd))
 			exit(1);
-		close(fd);
 	}
+
+	/*
+	 * If we opened a file, close it now even if there was an error
+	 * which made us decide not to write the archive body.
+	 */
+	if (fd >= 0)
+		close(fd);
 }
 
 /*
@@ -849,14 +855,8 @@ write_entry(struct bsdtar *bsdtar, struct archive *a, const struct stat *st,
     const char *pathname, const char *accpath)
 {
 	struct archive_entry	*entry, *sparse_entry;
-	int			fd;
-#ifdef __linux
-	int			 r;
-	unsigned long		 stflags;
-#endif
 	static char		 linkbuffer[PATH_MAX+1];
 
-	fd = -1;
 	entry = archive_entry_new();
 
 	archive_entry_set_pathname(entry, pathname);
@@ -904,17 +904,25 @@ write_entry(struct bsdtar *bsdtar, struct archive *a, const struct stat *st,
 	archive_entry_set_uname(entry, lookup_uname(bsdtar, st->st_uid));
 	archive_entry_set_gname(entry, lookup_gname(bsdtar, st->st_gid));
 
-#ifdef HAVE_CHFLAGS
+#ifdef HAVE_STRUCT_STAT_ST_FLAGS
+	/* XXX TODO: archive_entry_copy_stat() should copy st_flags
+	 * on platforms that support it.  This should go away then. */
 	if (st->st_flags != 0)
 		archive_entry_set_fflags(entry, st->st_flags, 0);
 #endif
 
-#ifdef __linux
-	if ((S_ISREG(st->st_mode) || S_ISDIR(st->st_mode)) &&
-	    ((fd = open(accpath, O_RDONLY|O_NONBLOCK)) >= 0) &&
-	    ((r = ioctl(fd, EXT2_IOC_GETFLAGS, &stflags)), close(fd), (fd = -1), r) >= 0 &&
-	    stflags) {
-		archive_entry_set_fflags(entry, stflags, 0);
+#ifdef EXT2_IOC_GETFLAGS
+	/* XXX TODO: Find a way to merge this with the
+	 * flags fetch for nodump support earlier. */
+	if ((S_ISREG(st->st_mode) || S_ISDIR(st->st_mode))) {
+		int fd = open(accpath, O_RDONLY | O_NONBLOCK);
+		if (fd >= 0) {
+			unsigned long stflags;
+			int r = ioctl(fd, EXT2_IOC_GETFLAGS, &stflags);
+			close(fd);
+			if (r == 0 && stflags != 0)
+				archive_entry_set_fflags(entry, stflags, 0);
+		}
 	}
 #endif
 
@@ -935,8 +943,7 @@ write_entry(struct bsdtar *bsdtar, struct archive *a, const struct stat *st,
 	siginfo_printinfo(bsdtar, 0);
 
 	while (entry != NULL) {
-		write_entry_backend(bsdtar, a, entry, fd);
-		fd = -1;
+		write_entry_backend(bsdtar, a, entry);
 		archive_entry_free(entry);
 		entry = sparse_entry;
 		sparse_entry = NULL;
@@ -947,30 +954,26 @@ cleanup:
 		fprintf(stderr, "\n");
 
 abort:
-	if (fd >= 0)
-		close(fd);
-
-	archive_entry_free(entry);
+	if (entry != NULL)
+		archive_entry_free(entry);
 }
 
 
-/* Helper function to copy file to archive, with stack-allocated buffer. */
+/* Helper function to copy file to archive. */
 static int
-write_file_data(struct bsdtar *bsdtar, struct archive *a, int fd)
+write_file_data(struct bsdtar *bsdtar, struct archive *a,
+    struct archive_entry *entry, int fd)
 {
-	char	buff[64*1024];
 	ssize_t	bytes_read;
 	ssize_t	bytes_written;
 	off_t	progress = 0;
 
-	/* XXX TODO: Allocate buffer on heap and store pointer to
-	 * it in bsdtar structure; arrange cleanup as well. XXX */
-
-	bytes_read = read(fd, buff, sizeof(buff));
+	bytes_read = read(fd, bsdtar->buff, FILEDATABUFLEN);
 	while (bytes_read > 0) {
 		siginfo_printinfo(bsdtar, progress);
 
-		bytes_written = archive_write_data(a, buff, bytes_read);
+		bytes_written = archive_write_data(a, bsdtar->buff,
+		    FILEDATABUFLEN);
 		if (bytes_written < 0) {
 			/* Write failed; this is bad */
 			bsdtar_warnc(bsdtar, 0, "%s", archive_error_string(a));
@@ -979,11 +982,12 @@ write_file_data(struct bsdtar *bsdtar, struct archive *a, int fd)
 		if (bytes_written < bytes_read) {
 			/* Write was truncated; warn but continue. */
 			bsdtar_warnc(bsdtar, 0,
-			    "Truncated write; file may have grown while being archived.");
+			    "%s: Truncated write; file may have grown while being archived.",
+			    archive_entry_pathname(entry));
 			return (0);
 		}
 		progress += bytes_written;
-		bytes_read = read(fd, buff, sizeof(buff));
+		bytes_read = read(fd, bsdtar->buff, FILEDATABUFLEN);
 	}
 	return 0;
 }
